package ru.ulstu.extractor;

import org.eclipse.jgit.api.Git;
import org.eclipse.jgit.api.errors.GitAPIException;
import org.eclipse.jgit.diff.DiffFormatter;
import org.eclipse.jgit.internal.storage.file.FileRepository;
import org.eclipse.jgit.lib.Repository;
import org.eclipse.jgit.revwalk.RevCommit;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.nio.file.Path;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Optional;

import static org.apache.logging.log4j.util.Strings.isBlank;

@Service
public class GitRepositoryService {
    @Value("${extractor.custom-projects-dir}")
    private String customProjectsDir;

    public void cloneOrUpdateRepo(String url) throws GitAPIException, IOException {
        Git git;
        if (projectDirExists(getProjectDirectoryFile(url))) {
            Repository localRepo = new FileRepository(getProjectGitDirectory(url));
            git = new Git(localRepo);
            git.pull().call();
        } else {
            git = Git.cloneRepository()
                    .setURI(url)
                    .setDirectory(getProjectDirectoryFile(url))
                    .call();
        }
        Iterable<RevCommit> commits = git.log().call();
        //commits.forEach(c -> System.out.println(c.getFullMessage()));
    }

    private String getProjectDirectory(String url) {
        return (isBlank(customProjectsDir)
                ? System.getProperty("java.io.tmpdir")
                : customProjectsDir) + url.substring(url.lastIndexOf('/'));
    }

    private String getProjectGitDirectory(String url) {
        return getProjectDirectory(url) + "/.git";
    }

    private File getProjectDirectoryFile(String url) {
        return Path.of(getProjectDirectory(url))
                .toFile();
    }

    private boolean projectDirExists(File file) {
        return file.exists();
    }

    public List<Commit> getCommits(String url) throws GitAPIException, IOException {
        cloneOrUpdateRepo(url);
        Repository localRepo = new FileRepository(getProjectGitDirectory(url));
        Git git = new Git(localRepo);
        git.pull().call();
        List<RevCommit> commits = new ArrayList<>();
        //TODO: сделать преобразование в коллекцию "наших объектов"
        git.log().call().forEach(commits::add);

        List<Commit> list = new ArrayList<>();
        RevCommit prevCommit = null;
        for (RevCommit revCommit : commits) {
            Commit commit = new Commit(
                    revCommit.getFullMessage(),
                    revCommit.getAuthorIdent().getName(),
                    Date.from(Instant.ofEpochSecond(revCommit.getCommitTime())));
            if (prevCommit != null) {
                commit.setChanges(findDiffBetweenTwoRevisions(revCommit, prevCommit, localRepo));
            }
            list.add(commit);
            prevCommit = revCommit;
        }
        return list;
    }

    public Changes findDiffBetweenTwoRevisions(RevCommit laterCommit, RevCommit earlierCommit, Repository localRepo) {
        if (laterCommit == null || earlierCommit == null) {
            return null;
        }
        String output = null;
        try (ByteArrayOutputStream out = new ByteArrayOutputStream()) {
            DiffFormatter diffFormatter = new DiffFormatter(out);
            diffFormatter.setRepository(localRepo);
            diffFormatter.format(earlierCommit.getId(), laterCommit.getId());
            output = out.toString();
        } catch (IOException e) {
            throw new RuntimeException("Error occurred during diff computation. Message: " + e.getMessage());
        }
        return parseOutputDiff(output);
    }

    private Changes parseOutputDiff(String output) {
        Changes changes = new Changes();
        String[] strings = output.split("\n");
        for (String string : strings) {
            Optional<String> maybeFileName = getFileName(string);
            if (maybeFileName.isPresent()) {
                FileChange fileChange = new FileChange();
                fileChange.setFile(maybeFileName.get());
                /// вытащить другие изменения из коммита
                changes.getFileChanges().add(fileChange);
            }
        }
        return changes;
    }

    private Optional<String> getFileName(String commitString) {
        String startString = "diff --git a/";
        if (commitString.startsWith(startString)) {
            String name = commitString.substring(commitString.indexOf(startString) + startString.length(), commitString.indexOf("b/"));
            return Optional.of(name);
        }
        return Optional.empty();
    }
}
